from factory import Factory, Trait
from contextlib import contextmanager
from threading import local
import logging

# Registry for all factories
factories = {}
traits = {}
debug = False
local = local()

# Methods to setup and declare factories
@contextmanager
def factory(name, collection=None):
    """Declares a new named factory with the given attributes."""
    factory = Factory(collection, global_traits=traits)
    factories[name] = factory

    # Set the context for other methods
    local.working_factory = factory
    yield
    del local.working_factory


def get_factory(name):
    """Get a factory by name"""
    return factories[name]


def _get_active_factory():
    if not hasattr(local, 'working_factory'):
        raise FactoryContextException("Method must be called inside a 'with factory()' context.")
    return local.working_factory


def default(attrs, traits=[]):
    _get_active_factory().default(attrs, traits)


def document(name, attrs=None, parent=None, traits=[]):
    _get_active_factory().document(name, attrs, parent, traits)


def trait(name, attrs, parent=None):
    if hasattr(local, 'working_factory'):
        _get_active_factory().trait(name, attrs, parent)
    else:
        traits[name] = Trait(attrs, parent)


def fragment(name, attrs=None, parent=None, traits=[]):
    _get_active_factory().fragment(name, attrs, parent, traits)


def embed(name):
    return _get_active_factory().embed(name)


# Methods to create document instances using factories
def create(factory_, document_=None, **overrides):
    """Creates and returns instance of the named document using the factory
    with which it was declared, utilising any provided attribute
    overrides, storing the instance in the database."""
    doc = factories[factory_].create(document_, **overrides)
    if debug:
        logging.debug("CREATED [%s]: %s, document=%s, overrides=%s", 
                      doc['_id'], factory_, document_, overrides)
    return doc


def build(factory_, document_=None, **overrides):
    """Builds and returns instance of the named document using the factory
    with which it was declared, utilising any provided attribute
    overrides, without storing the instance in the database."""
    return factories[factory_].build(document_, **overrides)


def build_list(count_, factory_, document_=None, **overrides):
    """Builds a list of `count_` instances of the named document using the 
    associated factory."""
    return [build(factory_, document_, **overrides) for x in range(count_)]


def create_list(count_, factory_, document_=None, **overrides):
    """Creates a list of `count_` instances of the named document using the 
    associated factory."""
    return [create(factory_, document_, **overrides) for x in range(count_)]


# Cleanup methods
def cleanup():
    """Cleans up all factory data generated since the process was started, 
    or since the last time this method was called."""
    for factory in factories.itervalues():
        factory.cleanup()

def reset():
    """Resets Monufacturer, removing all registered factories. Only really 
    here for testing purposes."""
    cleanup()
    factories.clear()
    traits.clear()


class FactoryContextException(Exception):
    pass



from types import FunctionType


def getitem(self, index, superclass):
    """Override __getitem__ implementation shared by both dynamic dicts
    and lists."""

    value = superclass.__getitem__(index)

    # If the value is a function, invoke it to get, set and return
    # the value.
    if isinstance(value, FunctionType):
        value = value(self)

    # If the value is an embedded dict, we need to wrap it in a
    # DynamicDict instance.
    if isinstance(value, dict):
        value = DynamicDict(value, head=self.head)

    # If the value is an embedded list, we need to wrap it in a
    # DynamicList instance.
    if isinstance(value, list):
        value = DynamicList(value, head=self.head)

    self[index] = value

    return superclass.__getitem__(index)


class DynamicList(list):
    """ A subclass of dict which checks whether a given index's value is a
    function, and if so return the result of calling that function.
    Note that each function is only called once."""

    def __init__(self, inner_list={}, head=None, *args, **kwargs):
        super(DynamicList, self).__init__(*args, **kwargs)
        self.head = self if not head else head
        self.extend(inner_list)

    def __getitem__(self, index):
        return getitem(self, index, super(DynamicList, self))

    def resolve(self):
        """"Resolves the dynamic list into a static list with
        static values."""
        
        out = []
        for i in range(len(self)):
            value = self[i]
            if isinstance(value, DynamicDict) or isinstance(value, DynamicList):
                out.append(value.resolve())
            else:
                out.append(value)

        return out

class DynamicDict(dict):
    """ A subclass of dict which checks whether a given key's value is a
    function, and if so return the result of calling that function.
    Note that each function is only called once."""

    def __init__(self, inner_dict={}, head=None, *args, **kwargs):
        super(DynamicDict, self).__init__(*args, **kwargs)
        self.head = self if not head else head
        self.update(inner_dict)

    def __getitem__(self, key):
        return getitem(self, key, super(DynamicDict, self))

    def resolve(self):
        """"Resolves the dynamic dictionary into a static dictionary with
        static values."""
        
        out = {}
        for key in self:
            value = self[key]

            if isinstance(value, DynamicDict) or isinstance(value, DynamicList):
                out[key] = value.resolve()
            else:
                out[key] = value

        return out



from dynamic import DynamicDict

class Document(object):
    def __init__(self, attrs, parent=None, traits=[]):
        self.attrs = attrs
        self.parent = parent
        self.traits = traits


class Trait(object):
    def __init__(self, attrs, parent=None):
        self.attrs = attrs
        self.parent = parent


class Fragment(object):
    def __init__(self, attrs, parent=None, traits=[]):
        self.attrs = attrs
        self.parent = parent
        self.traits = traits


class Factory(object):
    def __init__(self, collection=None, global_traits={}):
        self.collection = collection
        self.created_ids = []
        self.documents = {}
        self.traits = {}
        self.fragments = {}
        self.global_traits = global_traits


    def _apply_traits(self, doc, traits):
        for trait in traits:
            doc.update(self._build_trait(trait))

    def _build_fragment(self, name):
        fragment = self.fragments[name]
        if fragment.parent:
            spec = self._build_fragment(fragment.parent)
        else:
            spec = DynamicDict()

        if fragment.traits:
            self._apply_traits(spec, fragment.traits)

        spec.update(fragment.attrs)
        return spec

    def _build_trait(self, name):
        trait = self.traits.get(name) or self.global_traits.get(name)
        if trait.parent:
            spec = self._build_trait(trait.parent)
        else:
            spec = DynamicDict()

        spec.update(trait.attrs)
        return spec

    def _build_document(self, name):
        doc = self.documents[name]

        if doc.parent:
            spec = self._build_document(doc.parent)
        else:
            spec = DynamicDict()

        if doc.traits:
            self._apply_traits(spec, doc.traits)

        spec.update(doc.attrs)
        return spec

    def build(self, name_=None, **overrides):
        """Builds an instance of the document described by the attributes
        used to create this factory without actually persisting it to
        the database. Any overrides provided are used in preference to
        those attributes associated with the factory."""
        if not name_:
            name_ = "default"

        if name_ not in self.documents:
            raise NonExistentDocumentException(name_)

        spec = self._build_document(name_)

        spec.update(overrides)
        return spec.resolve()

    def create(self, name_=None, **overrides):
        """Builds an instance of the document using the same approach as
        `build` but also persists the document to the database."""
        if not self.collection:
            raise IOError("Cannot create an instance when no collection is provided.")

        doc = self.build(name_, **overrides)
        doc_id = self.collection.insert(doc, safe=True)
        self.created_ids.append(doc_id)
        return self.collection.find_one(doc_id)


    def cleanup(self):
        """Cleanup all instances created by this factory."""
        while len(self.created_ids) > 0:
            self.collection.remove(self.created_ids.pop(), safe=True)

    def default(self, attrs, traits=[]):
        """Sets the default document dict for the factory."""
        self.documents["default"] = Document(attrs, traits=traits)

    def document(self, name, attrs=None, parent=None, traits=[]):
        """Declares a named document type within the factory."""
        if name == 'default':
            raise FactoryDeclarationException("Cannot register a factory document with the name 'default'")

        self.documents[name] = Document(attrs or {}, parent, traits)

    def trait(self, name, attrs, parent=None):
        """Declares a reusable trait hash which can be referenced in
        documents."""
        self.traits[name] = Trait(attrs, parent)

    def fragment(self, name, attrs=None, parent=None, traits=[]):
        self.fragments[name] = Fragment(attrs or {}, parent, traits)

    def embed(self, name):
        def build(*args):
            return self._build_fragment(name)

        return build

class NonExistentDocumentException(Exception):
    """Raised when the caller attempts to access a non-existent
    document."""
    def __init__(self, name):
        self.name = name

    def __str__(self):
        return "Document declaration not found: \"%s\"" % self.name

class FactoryDeclarationException(Exception):
    """Raised when an error has been detected in the declaration of a
    factory."""
    pass



import monufacture
import string
import random
from datetime import datetime, timedelta
from bson.objectid import ObjectId

"""Contains setter functions designed to be used inline with
factory definitions to inject dynamic values into models as
and when they are built."""


class Sequence(object):
    def __init__(self):
        self.seq_num = 0

    def next(self):
        self.seq_num = self.seq_num + 1
        return self.seq_num


def sequence(fn=None):
    """Defines a sequential value for a factory attribute. On each successive
    invocation of this helper (i.e. when a new instance of a document is
    created by the enclosing factory) the given function is passed a
    sequentially incrementing number which should be used to return a dynamic
    value to be used on the model instance."""
    sequence = Sequence()

    if not fn:
        fn = lambda n: n

    def build(*args):
        return fn(sequence.next())
    return build


def dependent(fn):
    """Declares a value for a factory attribute which depends on other values
    in the document in order to be set. The given function or lambda is
    passed the document and should return the value to set."""
    def build(obj):
        return fn(obj)

    return build


def id_of(factory_, document_=None, **overrides):
    """Creates an instance using the given named factory and returns the
    ID of the persisted record. """
    def build(*args):
        # Flatten an function overrides
        instance_overrides = {}
        for key, value in overrides.iteritems():
            if callable(value):
                instance_overrides[key] = value(*args)
            else:
                instance_overrides[key] = value

        return monufacture.create(factory_, document_, **instance_overrides)["_id"]
    return build


def text(*args, **kwargs):
    """Aliases to random_text."""
    return random_text(*args, **kwargs)


def random_text(length=10, spaces=False, digits=False, upper=True, 
                lower=True, other_chars=[]):
    """Inserts some random text of the given length into the document."""

    # Build the char set we'll use
    char_set = []
    if upper:
        char_set += list(string.uppercase)
    if lower:
        char_set += list(string.lowercase)
    if spaces:
        char_set.append(" ")
    if digits:
        char_set += list(string.digits)
    char_set += other_chars

    def build(*args):
        return "".join([random.choice(char_set) for i in xrange(length)])
    return build


def dbref_to(factory, document=None):
    """Create a DBRef-type subdoc structure linking to a new instance of the
    given named factory type."""
    
    def build(*args):
        return {
            "$id": monufacture.create(factory, document)["_id"],
            "$ref": monufacture.get_factory(factory).collection.name
        }
    return build


def date(year=None, month=None, day=None, hour=None, minute=None, second=None, microsecond=None):
    """Returns a function to generate the current datetime for insertion into a document.
    Components of the date/time can be provided as overrides."""

    if year or month or day or hour or minute or second or microsecond:
        if not (year and month and day):
            raise ValueError("Either all components of a date must be provided or none of them.")

        def compact_args(**kwargs):
            out = {}
            out.update((k, v) for k, v in kwargs.iteritems() if v is not None)
            return out

        dt_args = compact_args(
            year=year, 
            month=month, 
            day=day, 
            hour=hour, 
            minute=minute, 
            second=second, 
            microsecond=microsecond)

        def build_specific(*args):
            return datetime(**dt_args)

        return build_specific

    def build_now(*args):
        return datetime.now()
    return build_now


def now():
    """Forwards to date(), allowing the current datetime to be inserted."""
    return date()

def _convert_years_months_to_days(timedeltas):
    """Converts any 'years' or 'months' values in a given timedeltas dict
    to equivalent values of days and adds them to the 'days' value.
    This is to work around a lack of support for months and years in
    Python's timedelta library.
    Assumes 30 days per month, 365 days per year.
    """
    if 'months' in timedeltas:
        timedeltas['days'] = timedeltas.setdefault('days', 0) + 30 * timedeltas['months']
        del timedeltas['months']
    if 'years' in timedeltas:
        timedeltas['days'] = timedeltas.setdefault('days', 0) + 365 * timedeltas['years']
        del timedeltas['years']

def ago(**kwargs):
    """Returns a function to generate a datetime a time delta in the past from the 
    time at which it is run."""
    def build(*args):
        _convert_years_months_to_days(kwargs)
        return datetime.now() - timedelta(**kwargs)

    return build


def from_now(**kwargs):
    """Returns a function to generate a datetime a time delta in the future from the
    time at which it is run."""
    def build(*args):
        _convert_years_months_to_days(kwargs)
        return datetime.now() + timedelta(**kwargs)

    return build


def list_of(fn, length):
    """Returns a function to generate a list of the given length,
    consisting of results of the given function"""
    def build(*args):
        return [fn(*args) for i in range(length)]
    return build


def object_id():
    """Returns a builder function which will insert a new ObjectId
    when the object is built."""
    def build(*args):
        return ObjectId()
    return build


def union(*fns):
    """Allows the list output of other helper functions to be unioned
    together in order to be set as a single attribute value. E.g.
    Unioning the output of n list_of helper calls."""
    def build(*args):
        out = []
        for fn in fns:
            out += fn(*args)
        return out
    return build


def one_of(*values):
    """Provides a function which returns one of the given values at
    random. Useful for getting a range of different but valid
    field values on a list of document instances."""
    def build(*args):
        return random.choice(values)
    return build


def random_number(a, b=None):
    """Inserts a random number in the given range into the document."""
    def build(*args):
        return random.randrange(a, b)
    return build


def number(*args, **kwargs):
    """Alias to random_number."""
    return random_number(*args, **kwargs)



import setuptools

setuptools.setup(
    name="Monufacture",
    version="0.2.13",
    author="Tom Leach",
    author_email="tom@gc.io",
    description="A lightweight factory framework for easily generating test data in MongoDB",
    license="BSD",
    keywords="mongo mongodb database testing factory pymongo",
    url="http://github.com/gamechanger/monufacture",
    packages=["monufacture"],
    long_description="Monufacture is a factory framework with an API designed to make " +
                     "it as easy as possible to generate valid test data in MongoDB. " +
                     "Inspired by the excellent factory_girl Ruby Gem.",
    install_requires=['pymongo'],
    tests_require=['mock', 'nose', 'freezegun']
)
